/******************************************************************************
Name: bridge_db.C

Description: This file holds all the functions that access the database, 
	     whether just to lookup information or to change the contents
	     of the database. It contains all of the functions in the
	     'GRID_DB_WORK_ITEM', 'GRID_APP_VERSION' and 'GRID_USER' classes.
	     
Authors: Jakob Gregor Pedersen & Christian Ulrik Sttrup.
******************************************************************************/

#ifndef _GRID_DB_BASE_
#define _GRID_DB_BASE_

#include "rsaeuro.h"

#define MAXIMUM_DB_INT 2147483647

class GRID_APP_VERSION {
public:
	void lookup(int appid);

	int version;
	char xml_doc[LARGE_BLOB_SIZE];
	char app_name[256];
};

class GRID_USER {
public:
	void lookup();
	
	int hostid;
	int userid;
	int teamid;
};

class GRID_DB_WORK_ITEM : public DB_WORK_ITEM {
public:
    	int add_job_to_grid_table(double grid_fpops_est);
	int add_result(GRID_USER gu, int app_version, time_t now);
	int get_grid_jobs(int limit, double grid_fpops_est, time_t now);
    	int get_possibly_finished_jobs(int limit);
	int handle_cleared_status();
    	int set_job_failed(int exit_status = 0);
	int set_job_finished();
    	int set_status_time(int new_status_time);
    
        char grid_jobid[256];
    	char xml_doc_in[LARGE_BLOB_SIZE];
    	int resultid;
	int report_deadline;
    		
private:
	void delete_files();
	void parse(MYSQL_ROW& r);
    	void parse_grid(MYSQL_ROW& r);
	
	int returned_grid_jobs;
	int returned_possibly_finished_jobs;
	int status_time;
};

#endif
