/******************************************************************************
Name: bridge_util.C

Description: This file holds a mixed collection of functions needed by the
	     BOINC->Grid bridge. Most of these functions create a file needed
	     by the job, this includes the jdl-file, the wrapper-script etc.
	     
Authors: Jakob Gregor Pedersen & Christian Ulrik Sttrup.

kommentarer:
hvorfor bogusserver i xml fillen?
hvorfor bruger du ikke string klassen det ville goere insert efter match meget
nemmere:
fstr
instr
str
str.insert(str.find(fstr)+fstr.size,instr);
******************************************************************************/

#include <sys/stat.h>
#include <fcntl.h>
#include "boinc_db.h"
#include "bridge_db.h"
#include "bridge_util.h"
#include "backend_lib.h"
#include "parse.h"
#include "sched_msgs.h"

#define JDL_FILE_SIZE 1024

/******************************************************************************
Name: create_account_xml

Function: Creates the account.xml file if not present. The file enables the 
	  modified BOINC core client to run jobs on the Grid.
	  
Called by: bridge_loop in bridge.C.
******************************************************************************/
int create_account_xml(){
	int retval;
	FILE *account_file;
	struct stat fileinfo;
	
	retval = stat("../grid/xml/account_bogusserver.xml", &fileinfo); //Does the account file already exist.
	if (!retval){ //The file already exist.
		return 0;		
	}
	
	//The file did not exist so create it and directories if needed.
	log_messages.printf(SCHED_MSG_LOG::DEBUG, "Creating 'account_bogusserver.xml'\n");
		
	mkdir("../grid", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	mkdir("../grid/xml", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	
	account_file = fopen("../grid/xml/account_bogusserver.xml","w");
	
	if(!account_file)
		return 1;
	
	retval = fputs("<account>\n"
			"\t<master_url>http://bogusserver</master_url>\n"
			"</account>\n"
			, account_file);
	if(retval==EOF){
		fclose(account_file);
		return 1;
	}
	
	fclose(account_file);
	return 0;
}


/******************************************************************************
Name: create_client_state_xml

Function: Creates a client_state.xml file for a job. The file enables the 
	  modified BOINC core client to run a single job on the Grid.
	  The function creates the file 'workunit-name'_client_state.xml in 
	  the ../grid/jobs directory, which if not present will be created too. 

Called by: create_files_for_job.
******************************************************************************/
int create_client_state_xml(GRID_DB_WORK_ITEM& gwi, GRID_APP_VERSION& gav){
	int retval;
	char *pos;
	char *pos2;
	char filename[256];
	char command_line[256];
	char env_vars[256];
	FILE *state_file;
	
	command_line[0] = 0;
	env_vars[0] = 0;

	//Create the directory.
	mkdir("../grid", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	mkdir("../grid/jobs", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	
	sprintf(filename, "../grid/jobs/%s_client_state.xml", gwi.wu.name);
	
	state_file = fopen(filename,"w");
	
	if(!state_file)
		return 1;
		
	// Write the beginning of the state-file. Nothing interesting here. 
	retval = fprintf(state_file,
			"<client_state>\n"
			"<host_info>\n"
			"</host_info>\n"
			"<time_stats>\n"
			"</time_stats>\n"
			"<net_stats>\n"
			"</net_stats>\n"
			"<project>\n"
    			"\t<master_url>http://bogusserver/</master_url>\n" //Maybe something else?
    			"\t<project_name>boinc@grid</project_name>\n"
   	 		"\t<share_size>0.000000</share_size>\n"
    			"\t<size>0.000000</size>\n"
    			"\t<user_name></user_name>\n"
    			"\t<team_name></team_name>\n"
    			"\t<email_hash></email_hash>\n"
    			"\t<cross_project_id></cross_project_id>\n"
			"\t<user_total_credit>0.000000</user_total_credit>\n"
    			"\t<user_expavg_credit>0.000000</user_expavg_credit>\n"
    			"\t<user_create_time>0</user_create_time>\n"
    			"\t<rpc_seqno>0</rpc_seqno>\n"
    			"\t<hostid>0</hostid>\n"
    			"\t<host_total_credit>0.000000</host_total_credit>\n"
    			"\t<host_expavg_credit>0.000000</host_expavg_credit>\n"
    			"\t<host_create_time>0</host_create_time>\n"
    			"\t<exp_avg_cpu>0.000000</exp_avg_cpu>\n"
    			"\t<exp_avg_mod_time>0.000000</exp_avg_mod_time>\n"
    			"\t<nrpc_failures>0</nrpc_failures>\n"
    			"\t<master_fetch_failures>0</master_fetch_failures>\n"
    			"\t<min_rpc_time>0</min_rpc_time>\n"
    			"\t<debt>0.000000</debt>\n"
    			"\t<master_url_fetch_pending/>\n"
    			"\t<sched_rpc_pending/>\n"
			"</project>\n"
			"<app>\n"
    			"\t<name>%s</name>\n"  
			"</app>\n", gav.app_name);
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Put the application information (files etc.) in the state-file (from the 'xml_doc' field in the 'app_version' table).
	
	retval = fputs(gav.xml_doc, state_file);
	if(retval==EOF){
		fclose(state_file);
		return 1;
	}
	
	//Getting information on the input files from the 'xml_doc' field in the 'workunit' table.
	
	retval = str_insert_after_match(gwi.wu.xml_doc, "<file_info>", "\n\t<status>1</status>", sizeof(gwi.wu.xml_doc));
	
	pos = strstr(gwi.wu.xml_doc, "<workunit>");
	
	if(!pos){
		fclose(state_file);
		return 1;
	}
	
	//Put the input file information in the state-file.
	retval = fwrite(gwi.wu.xml_doc, sizeof(char), pos-gwi.wu.xml_doc, state_file);
	
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Get command_line info
	pos = strstr(gwi.wu.xml_doc, "<command_line>");
	if(pos){
		pos2 = strstr(gwi.wu.xml_doc, "</command_line>");
		if (pos2){
			strncpy(command_line, pos+14, pos2-pos-14);
		}
	}
	
	//Get env_vars info
	pos = strstr(gwi.wu.xml_doc, "<env_vars>");
	if(pos){
		pos2 = strstr(gwi.wu.xml_doc, "</env_vars>");
		if (pos2){
			strncpy(env_vars, pos+10, pos2-pos-10);
		}
	}
	
	//Write information about the workunit to the state-file.	
	retval = fprintf(state_file,
			"<workunit>\n"
			"\t<name>%s</name>\n"
			"\t<app_name>%s</app_name>\n"
			"\t<version_num>%d</version_num>\n"
		        "\t<command_line>%s</command_line>\n"
    			"\t<env_vars>%s</env_vars>\n"
    			"\t<rsc_fpops_est>%f</rsc_fpops_est>\n" 
    			"\t<rsc_fpops_bound>%f</rsc_fpops_bound>\n"
    			"\t<rsc_memory_bound>%f</rsc_memory_bound>\n"
    			"\t<rsc_disk_bound>%f</rsc_disk_bound>\n" 			
			,gwi.wu.name, gav.app_name, gav.version, command_line, env_vars, gwi.wu.rsc_fpops_est, 
			gwi.wu.rsc_fpops_bound, gwi.wu.rsc_memory_bound, gwi.wu.rsc_disk_bound);
		
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Also put the file_ref information from the 'xml_doc' field in the 'workunit' table into the state-file.	
	pos = strstr(gwi.wu.xml_doc, "<file_ref>");
	
	if(!pos){
		fclose(state_file);
		return 1;
	}
	
	retval = fputs(pos, state_file);
	
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Get file_info's about the output files from the 'xml_doc_in' field from the 'result' table.
	
	pos = strstr(gwi.xml_doc_in, "<result>");
	
	if(!pos){
		fclose(state_file);
		return 1;
	}
	
	//Write the file_info's to the state-file.
	retval = fwrite(gwi.xml_doc_in, sizeof(char), pos-gwi.xml_doc_in, state_file);
		
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Write information about the result to the state-file.
	retval = fprintf(state_file,
			"<result>\n"
			"\t<name>%s_grid</name>\n"
			"\t<final_cpu_time>0.000000</final_cpu_time>\n"
			"\t<exit_status>0</exit_status>\n"
			"\t<state>2</state>\n"
		        "\t<wu_name>%s</wu_name>\n"
    			"\t<report_deadline>%d</report_deadline>\n    "
    			,gwi.wu.name, gwi.wu.name, MAXIMUM_DB_INT);
		
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Also put the file_ref's from the 'xml_doc_in' field from the 'result' table into the state-file.
	pos = strstr(gwi.xml_doc_in, "<file_ref>");
	
	if(!pos){
		fclose(state_file);
		return 1;
	}
	
	retval = fputs(pos, state_file);
	
	if(!retval){
		fclose(state_file);
		return 1;
	}
	
	//Write the end of the state-file. Nothing interesting here.
	retval = fputs("<active_task_set>\n"
			"</active_task_set>\n"
			"<platform_name>i686-pc-linux-gnu</platform_name>\n"
			"<core_client_major_version>4</core_client_major_version>\n"
			"<core_client_minor_version>3</core_client_minor_version>\n"
			"<cpu_sched_period>3600</cpu_sched_period>\n"
			"<cpu_sched_work_done_this_period>3600.000000</cpu_sched_work_done_this_period>\n"
			"<proxy_info>\n"
    			"\t<socks_version>5</socks_version>\n"
    			"\t<socks_server_name></socks_server_name>\n"
    			"\t<socks_server_port>80</socks_server_port>\n"
    			"\t<http_server_name></http_server_name>\n"
    			"\t<http_server_port>80</http_server_port>\n"
    			"\t<socks5_user_name></socks5_user_name>\n"
    			"\t<socks5_user_passwd></socks5_user_passwd>\n"
    			"\t<http_user_name></http_user_name>\n"
    			"\t<http_user_passwd></http_user_passwd>\n"
			"</proxy_info>\n"
			"</client_state>\n", state_file);
	if(retval==EOF){
		fclose(state_file);
		return 1;
	}
	
	retval = fclose(state_file);
	
	return retval;
}


/******************************************************************************
Name: create_jdl_file

Function: Creates a jdl-file for a job. The function creates the file 
	  'workunit-name'.jdl in the ../grid/jobs directory, which if not 
	  present will be created too. 

Called by: create_files_for_job.
******************************************************************************/
int create_jdl_file(GRID_DB_WORK_ITEM& gwi, GRID_APP_VERSION& gav, char* VO, int retrycount){
	char jdlfilename[256];
	char filename[256];
	char* pos;
	int retval;
	FILE *jdl_file;
	
	//Create the directory.
	mkdir("../grid", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	mkdir("../grid/jobs", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	
	sprintf(jdlfilename, "../grid/jobs/%s.jdl", gwi.wu.name);
	
	jdl_file = fopen(jdlfilename,"w");
	
	if(!jdl_file)
		return 1;
	
	//Write the beginning of the jdl-file.
	retval = fprintf(jdl_file,
		"[\n"
		"# Automatically generated by the BOINC_GRID-bridge.\n"
		"# Do not edit!\"\n\n"
		"Type = \"Job\";\n"
		"JobType = \"Normal\";\n"
		"Executable = \"%s_wrapper_script\";\n"  //Start by running the wrapper-script.
		"InputSandBox = {\"../grid/jobs/%s_wrapper_script\",\n" //Copy the wrapper-script to the Grid machine.
		      	"\t\"../grid/bin/boinc_client.gz\",\n" //Copy the modified core client (now zipped).
                	"\t\"../grid/xml/account_bogusserver.xml\",\n" //Copy the account-xml-file.
			"\t\"../grid/jobs/%s_client_state.xml\"", //And the state-file.
		gwi.wu.name, gwi.wu.name, gwi.wu.name);
		
	if(!retval){
		fclose(jdl_file);
		return 1;
	}
	
	//Get the names of the input files from the 'xml_doc' field form the 'workunit' table.	
	pos = strstr(gwi.wu.xml_doc, "<file_name>");
	
	//And make sure they all get copied to the Grid machine.	
	while (pos){
		if(!parse_str(pos, "<file_name>", filename, 256)){
			fclose(jdl_file);
			return 1;
		}
		retval = fprintf(jdl_file,
		",\n\t\"../download/%s\"",
		filename);
		if(!retval){
			fclose(jdl_file);
			return 1;
		}
		pos = strstr(pos+11, "<file_name>");
	}
	
	//Get the names of the files needed by the application from the 'xml_doc' field from the 'app_version' table.
	pos = strstr(gav.xml_doc, "<file_name>");
	
	//And make sure they all get copied to the Grid machine.
	while (pos){
		if(!parse_str(pos, "<file_name>", filename, 256)){
			fclose(jdl_file);
			return 1;
		}
		retval = fprintf(jdl_file,
		",\n\t\"../download/%s\"",
		filename);
		if(!retval){
			fclose(jdl_file);
			return 1;
		}
		pos = strstr(pos+11, "<file_name>");
	}	

	//Redirect stdout and stderr to files and make sure they get sent back, when the job finishes.
	retval = fprintf(jdl_file,
		"};\nStdOutput = \"%s_std.out\";\n"
		"StdError = \"%s_std.err\";\n"
		"OutputSandBox = {\"%s_std.out\",\n\t\"%s_std.err\"",
		gwi.wu.name, gwi.wu.name, gwi.wu.name, gwi.wu.name);
	
	if(!retval){
		fclose(jdl_file);
		return 1;
	}
	
	//Get the names of the output files from the 'xml_doc_in' field from the 'result' table.
	pos = strstr(gwi.xml_doc_in, "<file_name>");
	
	//And make sure they get sent back, when the job finishes.
	while (pos){
		if(!parse_str(pos, "<file_name>", filename, 256)){
			fclose(jdl_file);
			return 1;
		}
		retval = fprintf(jdl_file,
		",\n\t\"%s\"",
		filename);
		if(!retval){
			fclose(jdl_file);
			return 1;
		}
		pos = strstr(pos+11, "<file_name>");
	}	
	
	//Write the end of the jdl-file.	 
	retval = fprintf(jdl_file, 
		"};\n"
		"VirtualOrganisation = \"%s\";\n" //Use the VO set in the config.xml-file.
		"RetryCount = %d;\n" //Use the retrycount set in the config.xml-file.
		"Requirements = other.GlueCEStateStatus == \"Production\";\n"
		"Rank = - other.GlueCEStateEstimatedResponseTime;\n"
		"]\n"
		, VO, retrycount);
	
	if(!retval){
		fclose(jdl_file);
		return 1;
	}
	
	fclose(jdl_file);	
	return 0;
}


/******************************************************************************
Name: create_wrapper_script

Function: Creates a shell script to act as a wrapper for a job to be run on the 
	  Grid. The function creates the file 'workunit-name'_wrapper_script 
	  in the ../grid/jobs directory, which if not present will be 
	  created too. 

Called by: create_files_for_job.
******************************************************************************/
int create_wrapper_script(GRID_DB_WORK_ITEM& gwi, GRID_APP_VERSION& gav){
	char wrapperfilename[256];
	char filename[256];
	char *pos, *endpos, *execpos;
	int retval;
	FILE *wrapper_file;
	
	//Create directory
	mkdir("../grid", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	mkdir("../grid/jobs", O_CREAT|S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH);
	
	sprintf(wrapperfilename, "../grid/jobs/%s_wrapper_script", gwi.wu.name);
	
	wrapper_file = fopen(wrapperfilename,"w");
	
	if(!wrapper_file)
		return 1;
	
	//Make needed directories on the Grid machine.	
	retval = fputs("#! /bin/sh\n\n"
		"mkdir projects\n"
		"if [ $? -ne 0 ]\n"
        	"\tthen\n"
        	"\techo \"Could not create projects-dir!\"\n"
        	"\texit 1\n"
		"fi\n\n"	
		"mkdir projects/bogusserver\n"
		"if [ $? -ne 0 ]\n"
        	"\tthen\n"
        	"\techo \"Could not create bogusserver-dir!\"\n"
        	"\texit 2\n"
		"fi\n\n",
		wrapper_file);
	
	if(retval==EOF){
		fclose(wrapper_file);
		return 1;
	}
	
	//Get the names of the files needed by the application from the 'xml_doc' field from the 'app_version' table.
	pos = strstr(gav.xml_doc, "<file_name>");
	
	
	//And make sure they all get moved to the project directory.
	while(pos){
		if(!parse_str(pos, "<file_name>", filename, 256)){
			fclose(wrapper_file);
			return 1;
		}
		retval = fprintf(wrapper_file,
				"mv %s projects/bogusserver\n"
				"if [ $? -ne 0 ]\n"
        			"\tthen\n"
        			"\techo \"Could move %s!\"\n"
        			"\texit 3\n"
				"fi\n\n",
				filename, filename);
		if(!retval){
			fclose(wrapper_file);
			return 1;
		}
		pos = strstr(pos+11, "<file_name>");
	}
	
	//Get the names of the input files from the 'xml_doc' field from the 'workunit' table.
	pos = strstr(gwi.wu.xml_doc, "<file_name>");
	
	
	//And make sure they all get moved to the project directory.
	while(pos){
		if(!parse_str(pos, "<file_name>", filename, 256)){
			fclose(wrapper_file);
			return 1;
		}
		retval = fprintf(wrapper_file,
				"mv %s projects/bogusserver\n"
				"if [ $? -ne 0 ]\n"
        			"\tthen\n"
        			"\techo \"Could move %s!\"\n"
        			"\texit 4\n"
				"fi\n\n",
				filename, filename);
		if(!retval){
			fclose(wrapper_file);
			return 1;
		}
		pos = strstr(pos+11, "<file_name>");
	}

	//Get the names of the executable files.
	pos = strstr(gav.xml_doc, "<file_info>");
	endpos = strstr(gav.xml_doc, "</file_info>");
	execpos = strstr(gav.xml_doc, "<executable/>");
	
	//And make sure they all get execution rights on the Grid machine.	
	while(execpos){
		if((pos < execpos) && (execpos < endpos)){  
			pos = strstr(pos+11, "<name>");
			if(!parse_str(pos, "<name>", filename, 256)){
				fclose(wrapper_file);
				return 1;
			}
			retval = fprintf(wrapper_file,
					"chmod a+x projects/bogusserver/%s\n"
  					"if [ $? -ne 0 ]\n"
        				"\tthen\n"
        				"\techo \"Could not chmod %s!\"\n"
        				"\texit 5\n"
					"fi\n\n",
					filename, filename);
		}
		pos = strstr(endpos+11, "<file_name>");
		endpos = strstr(endpos+11, "<file_name>");
		execpos = strstr(endpos+11, "<executable/>");
	}
	
	//Rename the client_state-file.
	retval = fprintf(wrapper_file,
			"mv %s_client_state.xml client_state.xml\n"
			"if [ $? -ne 0 ]\n"
        		"\tthen\n"
        		"\techo \"Could move %s_client_state.xml!\"\n"
        		"\texit 6\n"
			"fi\n\n",
			gwi.wu.name, gwi.wu.name);
	
	if(!retval){
		fclose(wrapper_file);
		return 1;
	}
	
	//Make the modified core client executable and run it.
	retval = fprintf(wrapper_file,
			//Added to support zipping of the client.
			"gunzip boinc_client.gz\n"
			"if [ $? -ne 0 ]\n"
        		"\tthen\n"
        		"\techo \"Unzipping the Boinc client failed!\"\n"	
			"\texit 7\n"
			"fi\n\n"
			
			"chmod a+x boinc_client\n\n"
			"./boinc_client -skip_cpu_benchmarks -exit_when_idle\n"
			"if [ $? -ne 0 ]\n"
        		"\tthen\n"
        		"\techo \"Boinc_client failed!\"\n"	
			"\texit 8\n"
			"fi\n\n");
	
	if(!retval){
		fclose(wrapper_file);
		return 1;
	}
	
	//Get the names of the output files from the 'xml_doc_in' field from the 'result' table.
	pos = strstr(gwi.xml_doc_in, "<file_name>");
	
	//And make sure they all get moved to the working directory, so they can be sent back.	
	while (pos){
		if(!parse_str(pos, "<file_name>", filename, 256)){
			fclose(wrapper_file);
			return 1;
		}
		retval = fprintf(wrapper_file,
			"mv projects/bogusserver/%s .\n"
			"if [ $? -ne 0 ]\n"
        		"\tthen\n"
        		"\techo \"Could not move %s!\"\n"	
			"\texit 9\n"
			"fi\n\n",
			filename, filename);
		if(!retval){
			fclose(wrapper_file);
			return 1;
		}
		pos = strstr(pos+11, "<file_name>");
	}	
	
	//And finally exit.
	retval = fputs("exit 0\n", wrapper_file);
	
	if(retval==EOF){
		fclose(wrapper_file);
		return 1;
	}

	fclose(wrapper_file);
	return 0;
}


/******************************************************************************
Name: create_files_for_job

Function: Creates files needed to run a BOINC job on the Grid. These files are:
	  - Client_state.xml (file representing the state of a core client).
	  - jdl-file (file describing a job on the Grid). 
	  - Wrapper-script 
	    (a script creating a suitable environment for the job).

Called by: bridge_loop in bridge.C.
******************************************************************************/
int create_files_for_job(GRID_DB_WORK_ITEM& gwi, GRID_APP_VERSION& gav, char* VO, int retrycount){
	int retval;
			
	retval = create_client_state_xml(gwi, gav);
	if(retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to create client_state.xml for workunit %d.\n", gwi.wu.id);
		return 1;
		}
	retval = create_jdl_file(gwi, gav, VO, retrycount);
	if(retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to create jdl file for workunit %d.\n", gwi.wu.id);
		return 1;
		}
	retval = create_wrapper_script(gwi, gav);
	if(retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to create wrapper script for workunit %d.\n", gwi.wu.id);
		return 1;
		}
	return 0;
}


/******************************************************************************
Name: generate_result_xml_doc_in

Function: Generates the xml_doc_in for a new result. This is done by using the
	  appropriate template (found in the 'result_template_file' field of
	  the corresponding workunit) and processing it to insert urls, actual
	  filenames etc.

Called by: bridge_loop in bridge.C.
******************************************************************************/
int generate_result_xml_doc_in(GRID_DB_WORK_ITEM& gwi, R_RSA_PRIVATE_KEY key, char* upload_url){
	char base_outfile_name[256];
    	char rtfpath[256];
	int retval;
	
	sprintf(base_outfile_name, "%s_%s_", gwi.wu.name, "0"); //Naming the output files.
    	sprintf(rtfpath, "../%s", gwi.wu.result_template_file); //Getting the relative path of the template file.
	
	retval = read_filename(rtfpath, gwi.xml_doc_in, LARGE_BLOB_SIZE); //Reading the template file.
    	if (retval){
    		fprintf(stderr, 
		"Failed to read result template file '%s': %d\n",
		gwi.wu.result_template_file, retval);
		return retval;
    	}
	
	//Processing the template file to insert the private key, filenames and urls.
    	retval = process_result_template(gwi.xml_doc_in, key, base_outfile_name, upload_url); 
        
    	if (retval){
    		fprintf(stderr, 
		"Failed to process result template file '%s': %d\n",
		gwi.wu.result_template_file, retval);
		return retval;
    	}

    	return 0;
}


/******************************************************************************
Name: str_insert_after_match

Function: Searches a string for a given search string and insert a given
	  insert string after the search string everytime it is found. It
	  takes the size of the target string as an input to avoid overflowing.
	  It uses a temporary string of size 'LARGE_BLOB_SIZE' so the maximum
	  length of target string is also limited by this size.  If either the
	  target string size or the temporary string size will be exceeded the
	  function returns an error and applies no changes to the target 
	  string. If no errors occur the number of times the insert string
	  was inserted is returned.

Called by: create_client_state_xml and bridge_loop in bridge.C.
******************************************************************************/
int str_insert_after_match(char* target, const char* searchstr, const char* insertstr, int target_size){
	int len, numchars, insertstr_len, inserts=0;
	char temp[LARGE_BLOB_SIZE];
	char* pos;
	char* prevpos;
	
	len = strlen(searchstr); //Getting the length of the search string.
	if(!len)
		return 0;
	
	insertstr_len = strlen(insertstr); //Getting the length of the insert string.
	if(!len)
		return 0;
			
	temp[0]=0; //Initialising temp.
	
	prevpos = target; //the end of the 'previous' found search string - 'previous' since we haven't started seraching yet.
	
	pos = strstr(target, searchstr); //Find the first occurence of search string.
	
	while(pos){ //Keep going as long as we find more occurences of the search string.
		numchars = pos-prevpos+len; //The number of chars from the previous occurence of the search string.
		if((strlen(temp)+numchars+insertstr_len)>=sizeof(temp)) //Don't overflow.
			return -1;
			
		strncat(temp, prevpos, numchars); //cat the characters between the previous occurence and the new occurence to temp.
		 
		strcat(temp,insertstr); //cat the insert string to temp.
		
		prevpos = pos+len; //update previous position.
		pos = strstr(pos+1, searchstr); //Find the next occurence of search string.
		inserts++; //update inserts.
	}
	if((strlen(temp)+strlen(prevpos))>= sizeof(temp)) //Don't overflow.
		return -1;
				
	strcat(temp,prevpos); //Cat the characters between the last occurence and the end to temp.
	
	if(strlen(temp)>=(uint)target_size) //Don't overflow target.
		return -1;
	
	strcpy(target,temp); //Copy from temp to target.
	
	return inserts;
}


/******************************************************************************
Name: str_replace

Function: Replaces every occurence of a given character within a string with
	  another given character and returnes the number of replacements. 

Called by: set_job_failed in bridge_db.C.
******************************************************************************/
int str_replace(char* source, const char oldchar, const char newchar){
	int replacements = 0;
	int pos = 0;
	
	while(source[pos]){
		if (source[pos]==oldchar){
			source[pos]=newchar;
			replacements++;
		}
		pos++;
	}
	
	return replacements;
}  
